import matplotlib.pyplot as plt
import numpy as np
import time
import sys
import random


###### PARTIE 1 ######

def construction_mphf(set_kmer, n, gamma=2, nb_niveaux=3):
	
	"""
	Construit une fonction de hachage minimale parfaite (MPHF) pour un ensemble de k-mers.

	Parameters:
	set_kmer (set): Ensemble de k-mers.
	n (int): Taille de l'ensemble de k-mers.
	gamma (int): Facteur de réduction de la taille de la table de hachage. Default: 2.
	nb_niveaux (int): Nombre de niveaux de réduction de la taille de la table de hachage. Default: 3.

	Returns:
	mphf (list): Table de hachage minimale parfaite.

	Examples:
	>>> set_kmer = {str(i) for i in range(10)}
	>>> n = 10
	>>> mphf = construction_mphf(set_kmer, n)
	>>> len(mphf) == n
	True
	>>> all(0 <= mphf[i][1] < n for i in range(n))
	True
	>>> len(mphf) == n
	True
	"""
	# Initialisation
	# on copie le contenu de set_kmer dans set_kmer_courant
	set_kmer_courant = set_kmer.copy()
	# on crée un tableau vide
	tableaux = []
	# structure vide qui contiendra les collisions
	collision = set()
	# on parcours tous les niveaux
	for _ in range(nb_niveaux):
		# si la longueur du nombre de k_mers est positive == verifie que la liste n'est pas vide
		if len(set_kmer_courant) > 0:
			# on définit la longueur de la liste as l
			l = len(set_kmer_courant)
			# on initialise un tableau (une place pour chaque valeur) de taille double (gamma = 2) par rapport au nombre de valeurs
			tableau_principal = [-1] * (gamma * l)
			# on parcourt la liste de k_mers
			for kmer in set_kmer_courant:
				# hacher le k-mer (attention, hash() peut rendre des entiers signés, nous voulons des entiers positifs)
				h = abs(hash(kmer))
				# récupérer l'adresse / % = modulo -> ça garantit que la valeur est forcément ajoutée à un endroit du tableau
				adresse =  h%len(tableau_principal)
				# si le tableau principal est déjà rempli à l'adresse:
				if tableau_principal[adresse] != -1:
					# mettre le kmer dans collision()
					collision.add(kmer)
					#sinon, écrire le hash à l'adresse dans le tableau principal
				else : 
					tableau_principal[adresse] = h

			# Ici, tableaux correspond à la structure qui contient les tableaux des différents niveaux
			tableaux.append(tableau_principal) 
			# On incorpore à set_kmer_courant, l'ensemble des kmers n'ayant pas été ajouté au tableau principal à cause d'une collision
			set_kmer_courant = collision.copy() 
			# On réinitialise collision pour le niveau suivant
			collision = set()

	# Construction de la MPHF
	# On initialise une liste vide qui contiendra la table de hashage finale
	mphf = [] 
	# On initialise un tableau vide 
	grand_tableau = []
	# On parcours tous les tableaux dans tableaux (tableaux étant une liste de listes)
	for tableau in tableaux:
		# On fusionne tous les tableaux un même tableau
		grand_tableau.extend(tableau)

	# on initialise une variable rang qui contiendra les rangs des kmers
	rangs = []
	# On initialise le rang maximal par 0
	max_rang = 0
	# On parcours tous les kmers dans set_kmer
	for kmer in set_kmer:
		# hacher le kmer
		h2 = abs(hash(kmer))
		# si le hash est dans le grand_tableau
		if h2 in grand_tableau:
			# récupérer son index
			index = grand_tableau.index(h2)
			# récupérer son rang (utiliser la fonction count())
			c = grand_tableau[0:index].count(-1)
			rang = index - c
			# ajouter à la mphf [h, rang]
			mphf.append((h2, rang))
			# mettre à jour max_rang
			max_rang += 1

	# On parcourt la liste des kmers restant suite aux collisions 
	for kmer in set_kmer_courant: 
		# On aggrandit max_rang pour chaque valeur ajoutée
		max_rang += 1
		# On attribue une valeur de hashage au kmer
		h = abs(hash(kmer))
		# On l'ajoute à la fin du tableau final
		mphf.append([h, max_rang])

	# On trie la mphf par rang
	mphf = sorted(mphf, key = lambda x:x[1])

	return mphf


def get_hash_mphf(mphf, kmer):
	"""
	Calcule le hash d'un k-mer à l'aide d'une fonction de hachage minimale parfaite (MPHF).

	Parameters:
	mphf (list): Table de hachage minimale parfaite.
	kmer (str): K-mer à hasher.

	Returns:
	int: Hash du k-mer.

	Examples:
	>>> set_kmer = {str(i) for i in range(10)}
	>>> n = 10
	>>> mphf = construction_mphf(set_kmer, n)
	>>> kmer = "5"
	>>> hash_mphf = get_hash_mphf(mphf, kmer)
	>>> 0 <= hash_mphf < n
	True
	"""
	h = abs(hash(kmer))

	for h_kmer, rang in mphf:
		if h == h_kmer:
			return rang
	return -1
		
def create_hash_table(set_kmer, n):
    """			
	for _ in range(len(index)):
		c = grand_tableau.count(-1)chage.

    Returns:
    list: Table de hachage créée à partir des k-mers
    mphf: la mphf

    Examples:
    #>>> set_kmer = {'ATCG', 'TGCA', 'GCTA'}
    #>>> n = 10
    #>>> tableau = create_hash_table(set_kmer, n)
    #>>> len(tableau) == n
    #True
    #>>> all(kmer in tableau for kmer in set_kmer)
    #True
    """
    
    # créer la mphf pour les kmers
    mphf = construction_mphf(set_kmer, n)
    # initialiser un tableau de taille n (une liste)
    tableau = [None]*n
    # écrire les kmers aux adresses du tableau données par la mphf	
    for kmer in set_kmer:
        h = abs(hash(kmer))
        rang = get_hash_mphf(mphf, kmer)
        if 0 <= rang <n :
            tableau[rang] = kmer
    # retourner le tableau et la mphf
    return tableau, mphf



def generer_kmers(n, k):
	'''
	genere un set de n k-mers
	'''
	kmers = set()
	while len(kmers) < n:
		kmer = ''.join(random.choice('ATCG') for _ in range(k))
		kmers.add(kmer)
	return kmers


def compare_taille(n_max, fichier_sortie):
	n_values = []
	table_size = []
	dict_size = []
	k = 21
	
	for n in range(100, n_max, 1000):
		set_kmer = generer_kmers(n, k)
		tableau, mphf = create_hash_table(set_kmer,n)

		n_values.append(n)
		table_size.append(sys.getsizeof(tableau)+sys.getsizeof(mphf)) # pourquoi ici on ne mesure pas juste la taille en mémoire du tableau ? 
		dict_size.append(sys.getsizeof(set_kmer))

	plt.plot(n_values, table_size, label='Table avec MPHF')
	plt.plot(n_values, dict_size, label='Dict')
	plt.xlabel('n')
	plt.xticks(n_values, [str(x) for x in n_values], rotation=45)
	plt.ylabel('Taille (octets)')
	plt.title('Évolution de la taille de la table de hachage avec MPHF et du dict')
	plt.legend()
	plt.savefig(fichier_sortie)
	plt.close()

# dé-commenter quand vous êtes prêts, expliquer les résultats
compare_taille(10000,"mphf.png")

import doctest
doctest.testmod()

if __name__ == '__main__':
	set_kmer = {str(i) for i in range(10)}
	mphf = construction_mphf(set_kmer, 10)
	print(mphf)
	
	kmer = '5'
	print(abs(hash(kmer)))
	h= get_hash_mphf(mphf, kmer)
	print(h)
	print("hello world")
	print(create_hash_table(set_kmer, 10))